#include "verifier.h"

void frombytes(crypto_uint32 *x,const crypto_uint8 *s)
{
  crypto_uint32 c0;
  crypto_uint32 c1;
  crypto_uint32 c2;
  crypto_uint32 c3;
  x[0] = crypto_uint32(s[0]);
  x[0] += crypto_uint32(s[1]) << 8;
  x[0] += crypto_uint32(s[2]) << 16;
  x[0] += crypto_uint32(s[3]) << 24;
  x[1] = crypto_uint32(s[4]) << 6;
  x[1] += crypto_uint32(s[5]) << 14;
  x[1] += crypto_uint32(s[6]) << 22;
  x[2] = crypto_uint32(s[7]) << 4;
  x[2] += crypto_uint32(s[8]) << 12;
  x[2] += crypto_uint32(s[9]) << 20;
  x[3] = crypto_uint32(s[10]) << 2;
  x[3] += crypto_uint32(s[11]) << 10;
  x[3] += crypto_uint32(s[12]) << 18;
  x[4] = crypto_uint32(s[13]) << 0;
  x[4] += crypto_uint32(s[14]) << 8;
  x[4] += crypto_uint32(s[15]) << 16;
  c0 = x[0] >> 26; x[0] -= c0 << 26; x[1] += c0;
  c1 = x[1] >> 26; x[1] -= c1 << 26; x[2] += c1;
  c2 = x[2] >> 26; x[2] -= c2 << 26; x[3] += c2;
  c3 = x[3] >> 26; x[3] -= c3 << 26; x[4] += c3;
}

void mul(crypto_uint32 *z,const crypto_uint32 *x,const crypto_uint32 *y)
{
  crypto_uint64 x0 = x[0];
  crypto_uint64 x1 = x[1];
  crypto_uint64 x2 = x[2];
  crypto_uint64 x3 = x[3];
  crypto_uint64 x4 = x[4];
  crypto_uint64 y0 = y[0];
  crypto_uint64 y1 = y[1];
  crypto_uint64 y2 = y[2];
  crypto_uint64 y3 = y[3];
  crypto_uint64 y4 = y[4];
  crypto_uint64 z0 = x0 * y0;
  crypto_uint64 z1 = x0 * y1 + x1 * y0;
  crypto_uint64 z2 = x0 * y2 + x1 * y1 + x2 * y0;
  crypto_uint64 z3 = x0 * y3 + x1 * y2 + x2 * y1 + x3 * y0;
  crypto_uint64 z4 = x0 * y4 + x1 * y3 + x2 * y2 + x3 * y1 + x4 * y0;
  crypto_uint64 z5 = x1 * y4 + x2 * y3 + x3 * y2 + x4 * y1;
  crypto_uint64 z6 = x2 * y4 + x3 * y3 + x4 * y2;
  crypto_uint64 z7 = x3 * y4 + x4 * y3;
  crypto_uint64 z8 = x4 * y4;
  crypto_uint64 c0;
  crypto_uint64 c1;
  crypto_uint64 c2;
  crypto_uint64 c3;
  crypto_uint64 c4;
  z0 += z5 << 3;
  z1 += z6 << 3;
  z2 += z7 << 3;
  z3 += z8 << 3;
  c0 = z0 >> 26; z1 += c0; z0 -= c0 << 26;
  c1 = z1 >> 26; z2 += c1; z1 -= c1 << 26;
  c2 = z2 >> 26; z3 += c2; z2 -= c2 << 26;
  c3 = z3 >> 26; z4 += c3; z3 -= c3 << 26;
  c4 = z4 >> 23; z0 += c4; z4 -= c4 << 23;
  z[0] = z0;
  z[1] = z1;
  z[2] = z2;
  z[3] = z3;
  z[4] = z4;
}

void verifier_addlimbs(verifier_bigint *v,crypto_uint32 *x)
{
  for (int i = 0;i < 5;++i) verifier_addlimb(v,x[i],i * 26);
}

void doit(crypto_uint8 *s)
{
  crypto_uint32 x[5];
  crypto_uint32 z[5];
  verifier_modulus vp("170141183460469231731687303715884105727");

  verifier_bigint vs;
  for (int i = 0;i < 16;++i) verifier_addlimb(&vs,s[i],i * 8);

  frombytes(x,s);

  verifier_bigint vx;
  verifier_addlimbs(&vx,x);
  verifier_assertequal(&vs,&vx);

  mul(z,x,x);

  verifier_bigint vz;
  verifier_addlimbs(&vz,z);
  verifier_assertprodmod(&vz,&vx,&vx,vp);
}

main()
{
  crypto_uint8 s[16];

  for (int i = 0;i < 16;++i) s[i].inputunsignedchar();

  doit(s);
}
