#include "verifier.h"
#include "crypto_uint8.h"

int main()
{
  crypto_int32 p0 = 67108863;
  crypto_int32 p1 = 33554431;
  crypto_int32 p2 = 67108863;
  crypto_int32 p3 = 33554431;
  crypto_int32 p4 = 33554431;
  crypto_int32 h0;
  crypto_int32 h1;
  crypto_int32 h2;
  crypto_int32 h3;
  crypto_int32 h4;
  crypto_int32 carry0;
  crypto_int32 carry1;
  crypto_int32 carry2;
  crypto_int32 carry3;
  crypto_int32 carry4;
  crypto_int32 new0;
  crypto_int32 new1;
  crypto_int32 new2;
  crypto_int32 new3;
  crypto_int32 new4;
  crypto_int32 q;
  crypto_uint8 out[16];
  verifier_modulus vp("170141183460469231731687303715884105727");

  h0.inputint();
  h1.inputint();
  h2.inputint();
  h3.inputint();
  h4.inputint();

  verifier_bigint Horig;
  verifier_addlimb(&Horig,h0,0);
  verifier_addlimb(&Horig,h1,26);
  verifier_addlimb(&Horig,h2,51);
  verifier_addlimb(&Horig,h3,77);
  verifier_addlimb(&Horig,h4,102);

  verifier_assumerange(h0,"-300000000","300000000");
  verifier_assumerange(h1,"-300000000","300000000");
  verifier_assumerange(h2,"-300000000","300000000");
  verifier_assumerange(h3,"-300000000","300000000");
  verifier_assumerange(h4,"-300000000","300000000");

  carry4 = (h4 + 16777216) >> 25; h0 += carry4; h4 -= carry4 << 25;
  carry0 = (h0 + 33554432) >> 26; h1 += carry0; h0 -= carry0 << 26;
  carry1 = (h1 + 16777216) >> 25; h2 += carry1; h1 -= carry1 << 25;
  carry2 = (h2 + 33554432) >> 26; h3 += carry2; h2 -= carry2 << 26;
  carry3 = (h3 + 16777216) >> 25; h4 += carry3; h3 -= carry3 << 25;

  verifier_bigint H;
  verifier_addlimb(&H,h0,0);
  verifier_addlimb(&H,h1,26);
  verifier_addlimb(&H,h2,51);
  verifier_addlimb(&H,h3,77);
  verifier_addlimb(&H,h4,102);
  verifier_assertequalmod(&H,&Horig,vp);
  verifier_cutlimbs(&h0,1);
  verifier_cutlimbs(&h1,1);
  verifier_cutlimbs(&h2,1);
  verifier_cutlimbs(&h3,1);
  verifier_cutlimbs(&h4,1);

  int flagnegative = 0;

  if      (verifier_isnegative(h4)) flagnegative = 1;
  else if (verifier_ispositive(h4)) flagnegative = 0;
  else if (verifier_isnegative(h3)) flagnegative = 1;
  else if (verifier_ispositive(h3)) flagnegative = 0;
  else if (verifier_isnegative(h2)) flagnegative = 1;
  else if (verifier_ispositive(h2)) flagnegative = 0;
  else if (verifier_isnegative(h1)) flagnegative = 1;
  else if (verifier_ispositive(h1)) flagnegative = 0;
  else if (verifier_isnegative(h0)) flagnegative = 1;
  else flagnegative = 0;

  verifier_zero(&H);
  verifier_addlimb(&H,h0,0);
  verifier_addlimb(&H,h1,26);
  verifier_addlimb(&H,h2,51);
  verifier_addlimb(&H,h3,77);
  verifier_addlimb(&H,h4,102);
  // XXX: incorporate high-level H match as in testsmult.cpp

  q = h0 >> 30;
  q = (h1 + q) >> 30;
  q = (h2 + q) >> 30;
  q = (h3 + q) >> 30;
  q = (h4 + q) >> 30;

  verifier_cutlimbs(&q,1);

  h0 += q; h4 -= q << 25;
  carry0 = h0 >> 26; h1 += carry0; h0 -= carry0 << 26;
  carry1 = h1 >> 25; h2 += carry1; h1 -= carry1 << 25;
  carry2 = h2 >> 26; h3 += carry2; h2 -= carry2 << 26;
  carry3 = h3 >> 25; h4 += carry3; h3 -= carry3 << 25;

  crypto_int32 h03 = h0 >> 24;   crypto_int32 h03x = h0 - (h03 << 24);
  crypto_int32 h02 = h03x >> 16; crypto_int32 h02x = h03x - (h02 << 16);
  crypto_int32 h01 = h02x >> 8;  crypto_int32 h01x = h02x - (h01 << 8);
  crypto_int32 h00 = h01x;

  crypto_int32 h13 = h1 >> 22;   crypto_int32 h13x = h1 - (h13 << 22);
  crypto_int32 h12 = h13x >> 14; crypto_int32 h12x = h13x - (h12 << 14);
  crypto_int32 h11 = h12x >> 6;  crypto_int32 h11x = h12x - (h11 << 6);
  crypto_int32 h10 = h11x << 2;

  crypto_int32 h23 = h2 >> 21;   crypto_int32 h23x = h2 - (h23 << 21);
  crypto_int32 h22 = h23x >> 13; crypto_int32 h22x = h23x - (h22 << 13);
  crypto_int32 h21 = h22x >> 5;  crypto_int32 h21x = h22x - (h21 << 5);
  crypto_int32 h20 = h21x << 3;

  crypto_int32 h33 = h3 >> 19;   crypto_int32 h33x = h3 - (h33 << 19);
  crypto_int32 h32 = h33x >> 11; crypto_int32 h32x = h33x - (h32 << 11);
  crypto_int32 h31 = h32x >> 3;  crypto_int32 h31x = h32x - (h31 << 3);
  crypto_int32 h30 = h31x << 5;

  crypto_int32 h43 = h4 >> 18;   crypto_int32 h43x = h4 - (h43 << 18);
  crypto_int32 h42 = h43x >> 10; crypto_int32 h42x = h43x - (h42 << 10);
  crypto_int32 h41 = h42x >> 2;  crypto_int32 h41x = h42x - (h41 << 2);
  crypto_int32 h40 = h41x << 6;

  out[0] = h00;
  out[1] = h01;
  out[2] = h02;
  out[3] = h03 + h10;
  out[4] = h11;
  out[5] = h12;
  out[6] = h13 + h20;
  out[7] = h21;
  out[8] = h22;
  out[9] = h23 + h30;
  out[10] = h31;
  out[11] = h32;
  out[12] = h33 + h40;
  out[13] = h41;
  out[14] = h42;
  out[15] = h43;

  verifier_bigint Out;
  for (int i = 0;i < 16;++i) verifier_addlimb(&Out,out[i],i * 8);
  verifier_assertequalmod(&Out,&H,vp);
  verifier_bigint P;
  verifier_addlimb(&P,p0,0);
  verifier_addlimb(&P,p1,26);
  verifier_addlimb(&P,p2,51);
  verifier_addlimb(&P,p3,77);
  verifier_addlimb(&P,p4,102);
  if (flagnegative) verifier_assertsum(&Out,&H,&P);

  verifier_assertrange(&Out,"0","170141183460469231731687303715884105726");

  return 0;
}
