#include "crypto_int64.h"
#include "verifier.h"
#include "fe.h"

void fe_tobytes(crypto_uint8 *s,const crypto_int32 *h)
{
  crypto_int32 h0 = h[0];
  crypto_int32 h1 = h[1];
  crypto_int32 h2 = h[2];
  crypto_int32 h3 = h[3];
  crypto_int32 h4 = h[4];
  crypto_int32 h5 = h[5];
  crypto_int32 h6 = h[6];
  crypto_int32 h7 = h[7];
  crypto_int32 h8 = h[8];
  crypto_int32 h9 = h[9];
  crypto_int32 q;
  crypto_int32 carry0;
  crypto_int32 carry1;
  crypto_int32 carry2;
  crypto_int32 carry3;
  crypto_int32 carry4;
  crypto_int32 carry5;
  crypto_int32 carry6;
  crypto_int32 carry7;
  crypto_int32 carry8;
  crypto_int32 carry9;

  q = h0 >> 30;
  q = (h1 + q) >> 30;
  q = (h2 + q) >> 30;
  q = (h3 + q) >> 30;
  q = (h4 + q) >> 30;
  q = (h5 + q) >> 30;
  q = (h6 + q) >> 30;
  q = (h7 + q) >> 30;
  q = (h8 + q) >> 30;
  q = (h9 + q) >> 30;

  verifier_cutlimbs(&q,1);

  h0 += 19 * q; h9 -= q << 25;
  carry0 = h0 >> 26; h1 += carry0; h0 -= carry0 << 26;
  carry1 = h1 >> 25; h2 += carry1; h1 -= carry1 << 25;
  carry2 = h2 >> 26; h3 += carry2; h2 -= carry2 << 26;
  carry3 = h3 >> 25; h4 += carry3; h3 -= carry3 << 25;
  carry4 = h4 >> 26; h5 += carry4; h4 -= carry4 << 26;
  carry5 = h5 >> 25; h6 += carry5; h5 -= carry5 << 25;
  carry6 = h6 >> 26; h7 += carry6; h6 -= carry6 << 26;
  carry7 = h7 >> 25; h8 += carry7; h7 -= carry7 << 25;
  carry8 = h8 >> 26; h9 += carry8; h8 -= carry8 << 26;

  crypto_int32 h03 = h0 >> 24;   crypto_int32 h03x = h0 - (h03 << 24);
  crypto_int32 h02 = h03x >> 16; crypto_int32 h02x = h03x - (h02 << 16);
  crypto_int32 h01 = h02x >> 8;  crypto_int32 h01x = h02x - (h01 << 8);
  crypto_int32 h00 = h01x;

  crypto_int32 h13 = h1 >> 22;   crypto_int32 h13x = h1 - (h13 << 22);
  crypto_int32 h12 = h13x >> 14; crypto_int32 h12x = h13x - (h12 << 14);
  crypto_int32 h11 = h12x >> 6;  crypto_int32 h11x = h12x - (h11 << 6);
  crypto_int32 h10 = h11x << 2;

  crypto_int32 h23 = h2 >> 21;   crypto_int32 h23x = h2 - (h23 << 21);
  crypto_int32 h22 = h23x >> 13; crypto_int32 h22x = h23x - (h22 << 13);
  crypto_int32 h21 = h22x >> 5;  crypto_int32 h21x = h22x - (h21 << 5);
  crypto_int32 h20 = h21x << 3;

  crypto_int32 h33 = h3 >> 19;   crypto_int32 h33x = h3 - (h33 << 19);
  crypto_int32 h32 = h33x >> 11; crypto_int32 h32x = h33x - (h32 << 11);
  crypto_int32 h31 = h32x >> 3;  crypto_int32 h31x = h32x - (h31 << 3);
  crypto_int32 h30 = h31x << 5;

  crypto_int32 h43 = h4 >> 18;   crypto_int32 h43x = h4 - (h43 << 18);
  crypto_int32 h42 = h43x >> 10; crypto_int32 h42x = h43x - (h42 << 10);
  crypto_int32 h41 = h42x >> 2;  crypto_int32 h41x = h42x - (h41 << 2);
  crypto_int32 h40 = h41x << 6;

  crypto_int32 h53 = h5 >> 24;   crypto_int32 h53x = h5 - (h53 << 24);
  crypto_int32 h52 = h53x >> 16; crypto_int32 h52x = h53x - (h52 << 16);
  crypto_int32 h51 = h52x >> 8;  crypto_int32 h51x = h52x - (h51 << 8);
  crypto_int32 h50 = h51x;

  crypto_int32 h63 = h6 >> 23;   crypto_int32 h63x = h6 - (h63 << 23);
  crypto_int32 h62 = h63x >> 15; crypto_int32 h62x = h63x - (h62 << 15);
  crypto_int32 h61 = h62x >> 7;  crypto_int32 h61x = h62x - (h61 << 7);
  crypto_int32 h60 = h61x << 1;

  crypto_int32 h73 = h7 >> 21;   crypto_int32 h73x = h7 - (h73 << 21);
  crypto_int32 h72 = h73x >> 13; crypto_int32 h72x = h73x - (h72 << 13);
  crypto_int32 h71 = h72x >> 5;  crypto_int32 h71x = h72x - (h71 << 5);
  crypto_int32 h70 = h71x << 3;

  crypto_int32 h83 = h8 >> 20;   crypto_int32 h83x = h8 - (h83 << 20);
  crypto_int32 h82 = h83x >> 12; crypto_int32 h82x = h83x - (h82 << 12);
  crypto_int32 h81 = h82x >> 4;  crypto_int32 h81x = h82x - (h81 << 4);
  crypto_int32 h80 = h81x << 4;

  crypto_int32 h93 = h9 >> 18;   crypto_int32 h93x = h9 - (h93 << 18);
  crypto_int32 h92 = h93x >> 10; crypto_int32 h92x = h93x - (h92 << 10);
  crypto_int32 h91 = h92x >> 2;  crypto_int32 h91x = h92x - (h91 << 2);
  crypto_int32 h90 = h91x << 6;

  s[0] = h00;
  s[1] = h01;
  s[2] = h02;
  s[3] = h03 + h10;
  s[4] = h11;
  s[5] = h12;
  s[6] = h13 + h20;
  s[7] = h21;
  s[8] = h22;
  s[9] = h23 + h30;
  s[10] = h31;
  s[11] = h32;
  s[12] = h33 + h40;
  s[13] = h41;
  s[14] = h42;
  s[15] = h43;
  s[16] = h50;
  s[17] = h51;
  s[18] = h52;
  s[19] = h53 + h60;
  s[20] = h61;
  s[21] = h62;
  s[22] = h63 + h70;
  s[23] = h71;
  s[24] = h72;
  s[25] = h73 + h80;
  s[26] = h81;
  s[27] = h82;
  s[28] = h83 + h90;
  s[29] = h91;
  s[30] = h92;
  s[31] = h93;
}
