#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <iostream>
using namespace std;
#include "computation.h"
#include "verifier.h"
#include "crypto_uint64.h"

static int suppresshighlevel = 0;

void verifier_cutlimbs(const crypto_uint32 *u,long long n)
{
  while (n > 0) { computation_cut(u->id); --n; ++u; }
}

void verifier_cutlimbs(const crypto_int32 *u,long long n)
{
  while (n > 0) { computation_cut(u->id); --n; ++u; }
}

void verifier_zero(verifier_bigint *v)
{
  v->id = 0;
}

void verifier_addlimb(verifier_bigint *v,crypto_int32 u,long long e)
{
  long long shifted = computation_leftshift(u.id,e);
  v->id = computation_add(v->id,shifted);
}

void verifier_addlimb(verifier_bigint *v,crypto_int64 u,long long e)
{
  long long shifted = computation_leftshift(u.id,e);
  v->id = computation_add(v->id,shifted);
}

void verifier_addlimb(verifier_bigint *v,crypto_uint8 u,long long e)
{
  long long shifted = computation_leftshift(u.id,e);
  v->id = computation_add(v->id,shifted);
}

void verifier_addlimb(verifier_bigint *v,crypto_uint64 u,long long e)
{
  long long shifted = computation_leftshift(u.id,e);
  v->id = computation_add(v->id,shifted);
}

void verifier_addlimb(verifier_bigint *v,crypto_uint32 u,long long e)
{
  long long shifted = computation_leftshift(u.id,e);
  v->id = computation_add(v->id,shifted);
}

void verifier_limbs_10_255(verifier_bigint *v,const crypto_int32 *u)
{
  verifier_zero(v);
  verifier_addlimb(v,u[0],0);
  verifier_addlimb(v,u[1],26);
  verifier_addlimb(v,u[2],51);
  verifier_addlimb(v,u[3],77);
  verifier_addlimb(v,u[4],102);
  verifier_addlimb(v,u[5],128);
  verifier_addlimb(v,u[6],153);
  verifier_addlimb(v,u[7],179);
  verifier_addlimb(v,u[8],204);
  verifier_addlimb(v,u[9],230);
}

void verifier_assertxor(crypto_uint8 c,crypto_uint8 a,crypto_uint8 b)
{
  if (b.id == 0) {
    if (c.id != a.id) cout << "assert False # verifier_assertxor\n";
    return;
  }
  if (!suppresshighlevel) cout << "# highlevel v" << c.id << " = v" << a.id << " xor v" << b.id << "\n";
  computation_assertxor(c.id,a.id,b.id);
}

void verifier_assertsum(verifier_bigint *c,verifier_bigint *a,verifier_bigint *b)
{
  if (!suppresshighlevel) cout << "# highlevel v" << c->id << " = v" << a->id << " + v" << b->id << "\n";
  computation_assertequal(c->id,computation_add(a->id,b->id));
}

void verifier_assertdiff(verifier_bigint *c,verifier_bigint *a,verifier_bigint *b)
{
  if (!suppresshighlevel) cout << "# highlevel v" << c->id << " = v" << a->id << " - v" << b->id << "\n";
  computation_assertequal(c->id,computation_sub(a->id,b->id));
}

void verifier_assertprod(verifier_bigint *c,verifier_bigint *a,verifier_bigint *b)
{
  if (!suppresshighlevel) cout << "# highlevel v" << c->id << " = v" << a->id << " * v" << b->id << "\n";
  computation_assertequal(c->id,computation_mul(a->id,b->id));
}

void verifier_assumemin(computation_value u,const char *min)
{
  computation_assumemin(u.id,mpz_class(min));
}

void verifier_assumemax(computation_value u,const char *max)
{
  computation_assumemax(u.id,mpz_class(max));
}

void verifier_assumerange(computation_value u,const char *min,const char *max)
{
  computation_assumerange(u.id,mpz_class(min),mpz_class(max));
}

void verifier_assertrange(computation_value u,const char *min,const char *max)
{
  computation_assertrange(u.id,mpz_class(min),mpz_class(max));
}

void verifier_assertrange(verifier_bigint *c,const char *min,const char *max)
{
  computation_assertrange(c->id,mpz_class(min),mpz_class(max));
}

void verifier_assertconstant(verifier_bigint *c,mpz_class a)
{
  if (!suppresshighlevel) cout << "# highlevel v" << c->id << " = " << a << "\n";
  computation_assertconstant(c->id,a);
}

void verifier_assertequalmod(verifier_bigint *c,verifier_bigint *a,verifier_modulus mod)
{
  if (!suppresshighlevel) cout << "# highlevel v" << c->id << " = v" << a->id << " mod " << mod << "\n";
  computation_assertequalmod(c->id,a->id,mod);
}

void verifier_assertequal(verifier_bigint *c,verifier_bigint *a)
{
  if (!suppresshighlevel) cout << "# highlevel v" << c->id << " = v" << a->id << "\n";
  computation_assertequal(c->id,a->id);
}

void verifier_assertsummod(verifier_bigint *c,verifier_bigint *a,verifier_bigint *b,verifier_modulus mod)
{
  if (!suppresshighlevel) cout << "# highlevel v" << c->id << " = v" << a->id << " + v" << b->id << " mod " << mod << "\n";
  computation_assertequalmod(c->id,computation_add(a->id,b->id), mod);
}

void verifier_assertdiffmod(verifier_bigint *c,verifier_bigint *a,verifier_bigint *b,verifier_modulus mod)
{
  if (!suppresshighlevel) cout << "# highlevel v" << c->id << " = v" << a->id << " - v" << b->id << " mod " << mod << "\n";
  computation_assertequalmod(c->id,computation_sub(a->id,b->id), mod);
}

void verifier_assertprodmod(verifier_bigint *c,verifier_bigint *a,verifier_bigint *b,verifier_modulus mod)
{
  if (!suppresshighlevel) cout << "# highlevel v" << c->id << " = v" << a->id << " * v" << b->id << " mod " << mod << "\n";
  computation_assertequalmod(c->id,computation_mul(a->id,b->id), mod);
}

void verifier_assertcmov(verifier_bigint *x,verifier_bigint *a,verifier_bigint *b,crypto_uint8 c)
{
  if (!suppresshighlevel) cout << "# highlevel v" << x->id << " = v" << c.id << " ? v" << a->id << " : v" << b->id << "\n";
  computation_assertcmov(x->id,a->id,b->id,c.id);
}

void verifier_assertcswap(verifier_bigint *x,verifier_bigint *y,verifier_bigint *a,verifier_bigint *b,crypto_uint8 c)
{
  if (!suppresshighlevel) cout << "# highlevel v" << x->id << " = v" << c.id << " ? v" << a->id << " : v" << b->id << "\n";
  if (!suppresshighlevel) cout << "# highlevel v" << y->id << " = v" << c.id << " ? v" << b->id << " : v" << a->id << "\n";
  computation_assertcswap(x->id,y->id,a->id,b->id,c.id);
}

void verifier_assertbit(crypto_uint8 c,verifier_bigint *x,long long e)
{
  if (!suppresshighlevel) cout << "# highlevel v" << c.id << " = (v" << x->id << " >> " << e << ") & 1\n";
  computation_assertbit(c.id,x->id,e);
}

void verifier_input(verifier_bigint *x)
{
  if (!suppresshighlevel) cout << "# highlevel input v" << x->id << "\n";
}

void verifier_output(verifier_bigint *x)
{
  if (!suppresshighlevel) cout << "# highlevel output v" << x->id << "\n";
}

void verifier_highlevelcut(verifier_bigint *x)
{
  if (!suppresshighlevel) cout << "# highlevel cut v" << x->id << "\n";
}

int verifier_ispositive(computation_value u)
{
  if (computation_knownpositive(u.id)) return 1;
  if (computation_knownnonpositive(u.id)) return 0;
  cout << flush;
  pid_t child = fork();
  if (child == -1) {
    cout << "assert False\n";
    return 0;
  }
  if (child == 0) {
    cout << "# if v" << u.id << " > 0:\n";
    verifier_assumemin(u,"1");
    return 1;
  }
  waitpid(child,0,0);
  cout << "# elif v" << u.id << " <= 0:\n";
  suppresshighlevel = 1;
  verifier_assumemax(u,"0");
  return 0;
}

int verifier_isnegative(computation_value u)
{
  if (computation_knownnegative(u.id)) return 1;
  if (computation_knownnonnegative(u.id)) return 0;
  cout << flush;
  pid_t child = fork();
  if (child == -1) {
    cout << "assert False\n";
    return 0;
  }
  if (child == 0) {
    cout << "# if v" << u.id << " < 0:\n";
    suppresshighlevel = 1;
    verifier_assumemax(u,"-1");
    return 1;
  }
  waitpid(child,0,0);
  cout << "# elif v" << u.id << " >= 0:\n";
  verifier_assumemin(u,"0");
  return 0;
}
