#ifndef verifier_h
#define verifier_h

#include <gmpxx.h>

#include "crypto_int32.h"
#include "crypto_int64.h"
#include "crypto_uint64.h"

struct verifier_bigint {
  long long id;

  verifier_bigint()
  {
    id = 0;
  }
} ;

typedef mpz_class verifier_modulus;

void verifier_zero(verifier_bigint *);
void verifier_addlimb(verifier_bigint *,crypto_int32,long long);
void verifier_addlimb(verifier_bigint *,crypto_int64,long long);
void verifier_addlimb(verifier_bigint *,crypto_uint8,long long);
void verifier_addlimb(verifier_bigint *,crypto_uint32,long long);
void verifier_addlimb(verifier_bigint *,crypto_uint64,long long);
void verifier_limbs_10_255(verifier_bigint *,const crypto_int32 *);
void verifier_cutlimbs(const crypto_int32 *,long long);
void verifier_cutlimbs(const crypto_uint32 *,long long);

void verifier_assumemin(computation_value,const char *);
void verifier_assumemax(computation_value,const char *);
void verifier_assumerange(computation_value,const char *,const char *);
void verifier_assertrange(computation_value,const char *,const char *);
void verifier_assertrange(verifier_bigint *,const char *,const char *);
void verifier_assertconstant(verifier_bigint *,mpz_class);

void verifier_assertxor(crypto_uint8,crypto_uint8,crypto_uint8);
void verifier_assertsum(verifier_bigint *,verifier_bigint *,verifier_bigint *);
void verifier_assertdiff(verifier_bigint *,verifier_bigint *,verifier_bigint *);
void verifier_assertprod(verifier_bigint *,verifier_bigint *,verifier_bigint *);

void verifier_assertequal(verifier_bigint *,verifier_bigint *);
void verifier_assertequalmod(verifier_bigint *,verifier_bigint *,verifier_modulus);
void verifier_assertsummod(verifier_bigint *,verifier_bigint *,verifier_bigint *,verifier_modulus);
void verifier_assertdiffmod(verifier_bigint *,verifier_bigint *,verifier_bigint *,verifier_modulus);
void verifier_assertprodmod(verifier_bigint *,verifier_bigint *,verifier_bigint *,verifier_modulus);

void verifier_assertcmov(verifier_bigint *,verifier_bigint *,verifier_bigint *,crypto_uint8);
void verifier_assertcswap(verifier_bigint *,verifier_bigint *,verifier_bigint *,verifier_bigint *,crypto_uint8);
void verifier_assertbit(crypto_uint8,verifier_bigint *,long long);

void verifier_input(verifier_bigint *);
void verifier_highlevelcut(verifier_bigint *);
void verifier_output(verifier_bigint *);

int verifier_ispositive(computation_value);
int verifier_isnegative(computation_value);

#endif
